#ifndef POLYMAKE_YUE_KSPANS_H
#define POLYMAKE_YUE_KSPANS_H

#include "polymake/graph/Lattice.h"
#include "polymake/graph/BasicLatticeTypes.h"
#include "polymake/polytope/canonicalize.h"
#include "polymake/polytope/cdd_interface.h"

namespace polymake { namespace polytope {

   using graph::Lattice;
   using graph::lattice::Nonsequential;
   using graph::lattice::BasicDecoration;
   using graph::lattice::FaceIndexingData;
   using polymake::polytope::cdd_interface::solver;

   class trivial_closure_iterator {
      public:
         typedef Set<int> ClosureData;
         typedef std::forward_iterator_tag iterator_category;
         typedef ClosureData value_type;
         typedef const value_type& reference;
         typedef const value_type* pointer;
         typedef ptrdiff_t difference_type;

         trivial_closure_iterator(const ClosureData& input, int size) {
            for(auto rest = entire(sequence(0,size) - input); !rest.at_end(); ++rest)
               data.push_back( input + *rest);
            it = entire(data);
         }

         reference operator* () const { return it.operator*(); }
         pointer operator->() const { return it.operator->(); }
         trivial_closure_iterator& operator++ () { ++it; return *this; }
         const trivial_closure_iterator operator++ (int) { trivial_closure_iterator copy = *this; operator++(); return copy; }
         bool at_end() const { return it.at_end(); }



      protected:
         std::list<ClosureData> data;
         typename Entire<std::list<ClosureData> >::const_iterator it;
   };

   template <typename Decoration = BasicDecoration>
      class YueClosureOperator {
         public:
            typedef Set<int> ClosureData;

            YueClosureOperator(int size) : total_size(size) {}

            const ClosureData closure_of_empty_set() const {
               return Set<int>();
            }

            trivial_closure_iterator get_closure_iterator(const ClosureData& face) const {
               if(face.size() > max_size_observed) {
                  max_size_observed = face.size();
                  cout << endl << "Reached level " << max_size_observed << endl;
               }
               cerr << ".";
               return trivial_closure_iterator(face, total_size);
            }

            const ClosureData compute_closure_data(const Decoration& face) const {
               return face.face;
            }

            FaceIndexingData get_indexing_data(const ClosureData& data) {
               int& fi = face_index_map[data];
					return FaceIndexingData(fi, fi == -1, fi == -2);
            }

         protected:
            FaceMap<> face_index_map;
            const int total_size;
            mutable int max_size_observed = 0;
      };


   template <typename Decoration>
      class KSpanCut {
         protected:
            const Matrix<Rational> points;
            const int max_number;
            const Matrix<Rational> additional_vertices;
            mutable solver<Rational> sv;

         public:
            KSpanCut(const Matrix<Rational>& points,const Matrix<Rational>& additional_vertices, int max_k) : points(points), max_number(max_k), additional_vertices(additional_vertices), sv() {}
            inline bool operator()(const Decoration& data) const {
               if(data.face.size() > max_number) return false;
               Matrix<Rational> total_matrix = points.minor(data.face,All) / additional_vertices;
               Set<int> vertices = sv.canonicalize(total_matrix, Matrix<Rational>(), 1).first;
               return vertices.size() == data.face.size() + additional_vertices.rows();
            }

      };

   class YueDecorator {
      public:
         typedef Set<int> ClosureData;

         YueDecorator(const int total_size, const int k) : total_set(sequence(0,k)), k(k) {}

         const BasicDecoration compute_initial_decoration(const ClosureData& face) const {
            return BasicDecoration(face,0);
         }
         const BasicDecoration compute_decoration(const ClosureData& face, const BasicDecoration& predecessor_data) const {
            return BasicDecoration(face, face.size());
         }
         const BasicDecoration compute_artificial_decoration(const NodeMap<Directed, BasicDecoration> &decor,
               const std::list<int>& max_nodes) const {
            return BasicDecoration(total_set, k+1);
         }
      protected:
         const Set<int> total_set;
         const int k;
   };


}}

#endif
